declare name "FreeAx - Distortion-Feedback Electric Guitar Synth - Separate UI version";
declare author "Julius Smith";
declare version "1.0"; // LAC-2008 version is 1.0
declare license "STK-4.3"; // Synthesis Tool Kit 4.3 (MIT style license)
declare copyright "Julius Smith";
declare reference "http://ccrma.stanford.edu/realsimple/faust_strings/";
declare requires "faust 0.9.9.3b";

import("music.lib");    // Define SR, delay
import("oscillator.lib"); // /l/fa/osc.lib (oscs for LFO)
import("filter.lib");   // smooth, ffcombfilter
import("effect.lib");   // cubicnl, levelfilter, speakerbp

//==================== GUI SPECIFICATION ================

ui = hgroup("FreeAx",
  vgroup("[1]Excitation",
    // standard MIDI voice parameters:
    // NOTE: The labels MUST be "freq", "gain", and "gate" for faust2pd
    freq = nentry("freq", 440, 20, 7040, 1),  // Hz
    gain = nentry("gain", 0.1, 0, 1, 0.01),    // 0 to 1
    gate = button("gate"),                    // 0 or 1

    // Additional parameters (MIDI "controllers"):

    // Normalized pick-position in [0,0.5] - MIDI 0x81
    beta = hslider("excitation position", 0.13, 0.02, 0.5, 0.01),
       // MIDI Control 0x81 often "highpass filter frequency"

    // Level (just an overall level setting) - MIDI 7
    level = hslider("level-linear", 0.5, 0, 1, 0.01),

    // Pitchbend - MIDI 128
    pbend = hslider("pitchbend-semitones", 0, -2, 2, 0.1) : smooth(0.999) : pow(2),

    // Vibrato frequency in Hz:
    vfreq = hslider("vibrato freq Hz", 5.5, 1, 10, 0.1),

    // Vibrato amplitude (percent) - MIDI 1
    vpct = hslider("vibrato-percent", 0.3, 0, 10, 0.1)
  ),

vgroup("[2]Strings",

    // Detuning of coupled modes:
    detune = hslider("coupling detuning percent", 1, 0, 10, 0.1),

    // String decay time in seconds:
    t60 = hslider("decaytime T60", 4, 0, 10, 0.01),  // -60db decay time (sec)

    // Normalized brightness in [0,1]:
    B = hslider("brightness", 0.5, 0, 1, 0.01), // 0-1
	// MIDI Controller 0x74 is often "brightness" 
	// (or VCF lowpass cutoff freq)
  ),

vgroup("[3]Speaker",

    // Distortion "drive" in [0,1+) (gain going into the distortion curve):
    D  = hslider("distortion drive",0.629,0,10,0.01),

    // Amplifier feedback gain and delay
    ampfbgain  = hslider("Amp feedback gain",0.001,0,0.02,0.001),
    ampfbfres  = hslider("Amp feedback first resonance Hz",100,40,2000,1),

    // Speaker "model"
    // Defaults 130 and 5000 Hz based loosely on Celestion 12" measurements
    sf1  = hslider("Speaker low-frequency cutoff Hz",130,20,1000,1),
    sf2  = hslider("Speaker high-frequency cutoff Hz",5000,1000,10000,1)

));

// Dynamic level specified as dB level desired at Nyquist limit:
// L = hslider("excitation dynamic_level", -10, -60, 0, 1) : db2linear;
L = 128.0*(gain-1.0) : db2linear; // max = 0, min = -127 dB

// Spatial "width" (not in original EKS, but only costs "one tap"):
W = hslider("center-panned spatial width", 0.5, 0, 1, 0.01);
A = hslider("pan angle", 0.5, 0, 1, 0.01);

//==================== SIGNAL PROCESSING ================

//----------------------- noiseburst -------------------------
// White noise burst (adapted from Faust's karplus.dsp example)
// Requires music.lib (for noise)
noiseburst(gate,P) = noise : *(gate : trigger(P))
with {
  diffgtz(x) = (x-x') > 0;
  decay(n,x) = x - (x>0)/n;
  release(n) = + ~ decay(n);
  trigger(n) = diffgtz : release(n) : > (0.0);
};

ifreq = freq * pbend * (1.0 + (0.01*vpct) * oscs(vfreq));
P = SR/ifreq; // instantaneous fundamental period in samples
excitation = noiseburst(gate,P) : *(0.3); // signal.lib
Pmax = 4096;  // maximum P (for delay-line allocation below)

ppdel = beta*P; // pick position delay
pickposfilter = ffcombfilter(Pmax,ppdel,-1); // defined in filter.lib

// see effect.lib for levelfilter:
filtered_excitation = excitation : pickposfilter : 
   levelfilter(L,freq);
// levelfilterN(4,L,freq);

//stringloop = (+ : d1) ~ dampingfilter; //single-plane case
stringloop(lfreq) = 
  (+ <: d1,d2 : + : *(0.5)) ~ dampingfilter //dual-plane case:
    //or, more generally,
    // stringloop = ( _,+ : ((d2 <: _,_),(d1 <: _,_)) 
    //     : (_, (bridgefilter <: _,_) ,_) 
    //     : +,+) ~ (*(g2),*(g1)) ;
with 
{
  ilfreq = lfreq * pbend * (1.0 + (0.01*vpct) * oscs(vfreq));

  P = SR/ilfreq; // instantaneous fundamental period in samples

  // Linear phase FIR3 damping filter:
  rho = pow(0.001,1.0/(lfreq*t60)); // multiplies loop-gain
  h0 = (1.0 + B)/2; h1 = (1.0 - B)/4;
  dampingfilter(x) = rho * (h0 * x' + h1*(x+x''));

  d1 = fdelay1(Pmax, P-2);
  d2 = delay(Pmax, P*(1.0 - 0.01*detune)-2);
};

//Note re. stringloop above: fdelay1 is adequate when when 
// brightness or dynamic level are sufficiently low.
// Use fdelay4, e.g., for very bright, metallic strings.

distortion = cubicnl(D,0); // see effect.lib

// Second output decorrelated somewhat for spatial diversity over imaging:
widthdelay = delay(Pmax,W*P/2); // Delay ONE of the stereo output signals

// Assumes an optionally spatialized mono signal, centrally panned:
stereopanner(A) = _,_ : *(1.0-A), *(A);

ampfbdelay = int(SR/ampfbfres); // no interpolation needed, usually
ampfb = *(ampfbgain) : delay(4096,ampfbdelay);  
// 4096 should be pow(2,ceil(SR/ampfbresmin));

strings = stringloop(freq); // one string
//
// Parallel fifths (Smoke on Water demo):
// strings = _ <: stringloop(freq), stringloop(1.5*freq) :> _;
// (note that pickposfilter should be different for each string)
//
// Minor 7th chords (blues accompaniment demo):
// strings = _ <: 
//   stringloop(freq), 
//   stringloop(pow(2,10/12)*freq),
//   stringloop(pow(2,13/12)*freq),
//   stringloop(pow(2,19/12)*freq) 
// :> _;

//process = filtered_excitation : 
// (+ : strings : distortion : speakerbp(sf1,sf2)) // effect.lib
//  ~ ampfb : *(level) <: _,_ : _,widthdelay : stereopanner(A);

// Clones reused sliders (which we don't want):
//
// process = hgroup("FreeAx",
// 	vgroup("[1]Excitation",filtered_excitation) : 
// 	(+ :
// 	vgroup("[2]Strings",strings) <: 
// 	vgroup("[3]Speaker",
// 	  vgroup("[4]Distortion",distortion) : 
// 	  vgroup("[5]Speaker",speakerbp(sf1,sf2)))
// 	)
// 	~ vgroup("[6]Feedback",ampfb) :
// 	vgroup("[7]Output",*(level) <: _,_ : _,widthdelay : stereopanner(A)));

// Separate UI to inhibit cloning of reused sliders:

process = ui : filtered_excitation : 
	(+ : strings : distortion : speakerbp(sf1,sf2))
	~ ampfb : *(level) <: _,_ : _,widthdelay : stereopanner(A);

	vgroup("[3]Speaker",
	  vgroup("[4]Distortion",distortion) : 
	  vgroup("[5]Speaker",speakerbp(sf1,sf2)))
	)
	~ vgroup("[6]Feedback",ampfb) :
	vgroup("[7]Output",*(level) <: _,_ : _,widthdelay : stereopanner(A)));
