
/* pd.q: basic support functions for the Pd-Q external */

/* $Id: pd.q,v 1.14 2007/10/06 00:33:28 agraef Exp $ */

/* This module is to be imported in your pdext.q scripts. It is intended as a
   grab bag of general utility functions and useful objects for the Pd-Q
   interface. Currently it defines a PdMsg type with common types of Pd
   messages, some convenience functions for setting up Pd objects and a few
   sample objects which are used in the q-help.pd help patch. */

def EDITOR = "emacs"; // editor for the `q source' object, edit as needed

/* Realtime scheduling. This seems to be necessary to ensure that secondary
   threads get their fair share of timeslices when running under Pd. Adapt the
   scheduling policy and priority as needed. */

def POL = SCHED_RR, PRIO = 2;

/****************************************************************************/

/* PdMsg type: Common types of Pd messages. In Q land, Pd messages are
   represented as algebraic constructor terms of the form `MSG X1 X2 ...',
   where MSG is the Pd message selector as a Q symbol or a string, and X1,
   X2, ...  are the Q representations of the message parameters. The PdMsg
   type declares some common Pd message selectors as constructor symbols in
   Q. This works like the MidiMsg type of Q's midi module, but note that in Pd
   the set of message selectors is not fixed, so you might wish to derive your
   own subtypes of PdMsg if you want to declare additional message selectors
   that you might need (this isn't strictly necessary, though, as Q allows you
   to use ad-hoc constructor symbols without explicitly declaring them). */

/* NOTES: The list message is actually variadic, but as we can't declare a
   function symbol with a varying number of curried arguments in Q, we treat
   this symbol as a one-argument constructor here. Also note that `float' and
   `list' shadow the corresponding Q built-ins, so if you need to access these
   after importing this module, you have to write them in their qualified
   forms, ::float and ::list. */

public type PdMsg = const

  /* builtin message selectors */
  bang, float X, symbol X, list X /* ... */,

  /* The following are not built into Pd, but appear to be useful to represent
     common types of Pd events. */

  /* MIDI-like sequencer messages */
  start, stop, continue, reset,

  /* hiatons, clock and tempo messages (timed streams) */
  delta T, clock, tempo MSPP,

  /* MIDI-related messages */
  note CHAN NOTE VEL /* DUR */, ctl CHAN CTL VAL,
  pgm CHAN PGM, bend CHAN VAL, touch CHAN VAL, polytouch CHAN NOTE VAL,

  /* key-related messages */
  key VAL, keyup VAL, keyname STATE NAME;

/****************************************************************************/

/* Support for variadic creation functions. `varobj (varargs F X1 X2 ... Xn)'
   gives you an object creation function F which receives all its arguments as
   a list. This is useful if an object has to process a varying number of
   creation arguments. For instance: varobj (varargs foo 1 2 3) = foo
   [1,2,3]. The `varobj' function is applied automatically by the Pd/Q
   interface when the object is created, but needs to be applied manually when
   running the object creation function outside Pd. */

public varobj X, varargs F;
private mkargs F Xs;

varobj (mkargs F Xs)	= F Xs;
varobj X		= X otherwise;
varargs F		= mkargs F [];
mkargs F Xs X		= mkargs F (append Xs X);

/* Support for variadic messages works in a similar fashion. E.g.: varmsg
   (list 1 2 3) = list [1,2,3]. This is useful if an object has to process
   messages with a varying number of arguments. */

public varmsg X;
private mkmsg X Xs;

varmsg X		= mkmsg X [];
mkmsg (F X) Xs		= mkmsg F [X|Xs];
mkmsg F Xs		= F Xs otherwise;

/* Helper function to parse an object description of the form:

	(F[,N[,M]][;X1,X2,...,Xn][;Y1,Y2,...,Ym])

   which denotes the object function F and, optionally, the desired number of
   inlets N and outlets M and additional "visible state parameters" X1, X2,
   ..., Xn and "invisible state parameters" Y1, Y2, ..., Ym.

   - Just a non-tuple F is used to denote an object function with 1 inlet and
     1 outlet and no state parameters.

   - A tuple (F,N,M) denotes an object function F with the given number of
     inlets and outlets and no state parameters. N and M are optional and
     default to 1.

   - A pair (F,N,M;X1,X2,...,Xn) denotes an object function F with the given
     number of inlets and outlets and visible state parameters X1, X2, ...,
     Xn. N and M are optional and default to 1.

   - A triple (F,N,M;X1,X2,...,Xn;Y1,Y2,...,Ym) denotes an object function F
     with the given number of inlets and outlets, visible state parameters X1,
     X2, ..., Xn and invisible state parameters Y1, Y2, ..., Ym. N and M are
     optional and default to 1.

   The parse_obj function returns the object description in the normalized
   form of a triple ((F,N,M),(X1,X2,...,Xn),(Y1,Y2,...,Ym)). */

public parse_obj ARGS;
private parse_fun ARGS;

parse_obj (F:Tuple,Xs:Tuple,Ys:Tuple)
			= ((F,N,M),Xs,Ys) where (F,N,M) = parse_fun F;
parse_obj (F:Tuple,Xs:Tuple)
			= ((F,N,M),Xs,()) where (F,N,M) = parse_fun F;
parse_obj (F:Tuple,)	= ((F,N,M),(),()) where (F,N,M) = parse_fun F;
parse_obj F		= ((F,N,M),(),()) where (F,N,M) = parse_fun F;

parse_fun (F,N:Int,M:Int)
			= (F,N,M);
parse_fun (F,N:Int)	= (F,N,1);
parse_fun (F,)		= (F,1,1);
parse_fun F		= (F,1,1) if not istuple F;

/****************************************************************************/

/* The external type used to represent Pd objects. */

public extern type PdObj;

/* Internals used to implement asynchronous messaging. These need to be public
   so that the Pd-Q plugin can access them, but normally shouldn't be used
   directly in applications. NOTE: All this stuff only works as advertized if
   the module is actually run inside Pd. */

/* The global PDOBJ variable is set to the current Pd object at object
   creation time; it is undefined otherwise. This variable can be used by
   object creation functions to gain access to the created object, which can
   then be stored away and later be passed to the functions below. */

public var PDOBJ;

/* The PDGLOB variable has as its value the global "script" object; this
   shouldn't normally be used, but is provided here for future extensions.
   XXXFIXME: For the life of me, I can't remember what this was supposed to be
   used for. Does it prevent the global object from being garbage-collected? 
   Can we send this object a 'script' message to load a new script? We'll have
   to find out someday... */

public var PDGLOB;

/* The pd_send function sends Pd messages to the outlets of a Pd object. The
   given Q expression X to be sent has the same format as the result returned
   by an object function, i.e., it can be just a message, a pair consisting of
   an outlet index and a message, or a tuple or list of such objects. This
   function can be used in secondary "worker" threads to return results of an
   object in an asynchronous fashion. See the `stream_obj' function below for
   an example. Note that in contrast to messages returned directly by the
   object functions there might be small latencies (usually less than 1ms)
   until the messages are actually processed (the pd_send function merely
   buffers the messages for later processing in the main Pd thread). */

public extern pd_send OBJ X;

/* Default rule for pd_send which just prints the object on standard output.
   This is useful if you test your pdext.q script outside Pd. You can also
   override this rule in your application. */

@-0x80000000
pd_send _ X		= printf "%s\n" (str X);
@0

/* The pd_time function returns the current (logical) Pd time (a double
   precision floating point number, measured in milliseconds). This value is
   updated at least once per millisecond and whenever a creation or object
   function is invoked, and can be used by Pd/Q objects to perform their own
   internal timing calculations. */

public extern pd_time OBJ;

/* The pd_wait function suspends execution until the specified Pd time
   arrives. This can be used by secondary threads to implement delays (never
   use this in the main thread as it might block Pd for an extended period of
   time). Currently only a single pd_wait can be scheduled for each Pd
   object. If the specified wakeup time has already expired then pd_wait will
   return immediately. */

public extern pd_wait OBJ T;

/****************************************************************************/

/* Convenience functions to create Pd objects. */

/* The obj function creates a Pd object with a number of "state variables"
   with given default values. It accepts an object description of the form
   (F[,N[,M]][;X1,X2,...,Xn][;Y1,Y2,...,Ym]) (see the description of parse_obj
   above), and yields a variadic object creation function which accepts an
   arbitrary number of creation arguments overriding the defaults given in the
   (X1,X2,...,Xn) tuple (if present). If the number of actual parameters Z1,
   ..., Zk is less than the number of "visible" state variables, then only the
   first k variables X1, ..., Xk will be overridden.  If the number of actual
   parameters exceeds the number of visible state variables then an exception
   will be raised.

   During processing, the object function F receives the state variables as
   additional arguments, i.e., F will be invoked as F X1 ... Xn Y1 ... Ym MSG,
   where MSG is the Pd message to be processed. Note that the "hidden" state
   variables Y1 ... Ym will always be passed through unchanged and cannot be
   overridden by object creation arguments.

   State variables can also be references to implement mutable state. In this
   case creation arguments will override the values of the references instead
   of the references themselves. */

public obj ARGS;
private create_obj ARGS, create_var X Y;

obj ARGS		= varargs (create_obj ARGS)
			    where ARGS:Tuple = parse_obj ARGS;

create_obj ((F,N,M),Xs,Ys) Zs
			= throw "too many parameters" if #Zs > #Xs;
			= (foldl ($) F Xs,N,M)
			    where Xs = ::list Xs, Ys = ::list Ys,
			      Xs = zipwith create_var Xs Zs++drop (#Zs) Xs++Ys;

create_var X:Ref Y	= put X Y || X;
create_var X Y		= Y otherwise;

/****************************************************************************/

/* Stream processing objects. The stream_obj function works like obj, but
   takes a stream processing function which maps an input to an output stream.
   It creates a Pd object which constructs the input stream from messages
   received on its inlet(s), passes the input stream to the given function (in
   addition to the state parameters, if any), and sends the elements of the
   resulting output stream to its outlet(s) as they become available. All
   processing is done lazily and asynchronously in realtime. Note that, as the
   results are actually produced in a secondary "worker" thread, you can *not*
   rely on outputs to be delivered immediately in response to the
   corresponding inputs, as is the case with standard Pd/Q objects. */

public stream_obj ARGS;
private create_stream_obj ARGS, process OBJ F;

stream_obj ARGS		= varargs (create_stream_obj ARGS)
			    where ARGS:Tuple = parse_obj ARGS;

create_stream_obj ((F,N,M),Xs,Ys) Zs
			= throw "too many parameters" if #Zs > #Xs;
			= (process OBJ (foldl ($) F Xs),N,M)
			    where OBJ = PDOBJ,
			      Xs = ::list Xs, Ys = ::list Ys,
			      Xs = zipwith create_var Xs Zs++drop (#Zs) Xs++Ys;

/* The basic stream processing routine. Our implementation employs two
   separate threads which are executed concurrently: the "reader" is executed
   in the Pd main thread, takes messages from the object's inlets and passes
   them on to the writer thread via a semaphore queue; the "writer" is
   executed in a secondary "worker" thread, builds the input stream from the
   received messages, applies the supplied stream processing function and
   traverses the resulting output stream, sending results to the object's
   outlets. */

private reader H IN MSG, writer F IN OBJ;

process OBJ F		= setsched H POL PRIO || reader H IN
			    where IN:Semaphore = semaphore,
			      H = thread (sleep 0.1 || writer F IN OBJ);

/* The reader first checks whether the writer is still alive and kicking. If
   so then it feeds the received input message to the IN queue and returns. If
   the writer is dead then it just performs a no-op instead. (You can also
   uncomment the second equation below if you rather want the reader to raise
   an exception if the writer thread has exited.) */

reader H IN MSG		= post IN MSG if active H;
//			= throw "reader: writer has exited";
			= () otherwise;

/* The writer reads the input stream from the IN queue, applies the stream
   processing function, and sends the resulting output stream back to the
   object's outlets. The input stream is memoized so that we do not
   accidentally read the same element twice. */

private read_stream IN, write_stream OBJ Xs;

writer F IN OBJ		= write_stream OBJ (F (read_stream IN));

read_stream IN		= {&(get IN)|&(read_stream IN)};

write_stream OBJ {}	= ();
write_stream OBJ {X|Xs}	= pd_send OBJ X || write_stream OBJ Xs;

/* Timed stream objects. Timed streams are represented as streams with special
   interspersed tokens ("hiatons") of the form delta T denoting a delay time
   of T milliseconds (integer or floating point value) between events. The
   tstream_obj works like stream_obj, but automatically inserts the
   appropriate hiatons into the realtime input stream, and translates hiatons
   in the output stream into the appropriate realtime delays, using the
   internal Pd clock for reference. */

public tstream_obj ARGS;
private create_tstream_obj ARGS, tprocess OBJ F;

private tstream_in OBJ T S, tstream_out OBJ T S;

tstream_obj ARGS	= varargs (create_tstream_obj ARGS)
			    where ARGS:Tuple = parse_obj ARGS;

create_tstream_obj ((F,N,M),Xs,Ys) Zs
			= throw "too many parameters" if #Zs > #Xs;
			= (tprocess OBJ (foldl ($) F Xs),N,M)
			    where OBJ = PDOBJ,
			      Xs = ::list Xs, Ys = ::list Ys,
			      Xs = zipwith create_var Xs Zs++drop (#Zs) Xs++Ys;

private treader T H IN OBJ MSG, twriter T F IN OBJ;

tprocess OBJ F		= setsched H POL PRIO || treader (ref T) H IN OBJ
			    where IN:Semaphore = semaphore, T = pd_time OBJ,
			      H = thread (sleep 0.1 || twriter T F IN OBJ);

treader T H IN OBJ MSG	= if T1>T0 then
			    (post IN (delta (T1-T0)) || put T T1) ||
			  post IN MSG
			    where T1 = pd_time OBJ, T0 = get T
			    if active H;
//			= throw "treader: twriter has exited";
			= () otherwise;

private twrite_stream OBJ T Xs;

twriter T F IN OBJ	= twrite_stream OBJ T (F (read_stream IN));

twrite_stream OBJ _ {}	= ();
twrite_stream OBJ T {delta DT|Xs}
			= pd_wait OBJ T1 || twrite_stream OBJ T1 Xs
			    if T1>T2 where T2 = pd_time OBJ, T1 = T+DT;
			= twrite_stream OBJ (T+DT) Xs;
twrite_stream OBJ T {X|Xs}
			= pd_send OBJ X || twrite_stream OBJ T Xs;
//twrite_stream OBJ T Xs	= printf "bad: %s\n" (str Xs);

/****************************************************************************/

/* Basic examples. Some of these are used in the q-help.pd help patch so you
   shouldn't remove them. */

/* [q source] -- open the Q script for the current patch (pdext.q by default)
   in a text editor */

public source;
private edit X;

from system import spawn;

source			= (edit,1,0);
edit _			= spawn EDITOR [EDITOR,which ()];

/* [q echo] -- print the Q representation of a message on stdout. This example
   illustrates the most basic kind of Q object with a single inlet/outlet pair
   and no internal state. */

public echo X;

echo X			= printf "%s\n" S || S where S = str X;

/* Stateful counter. This example shows how the obj function can be employed
   to program a much more elaborate object with internal state, a variable
   number of creation args from which the state variables are initialized, and
   multiple inlets and outlets. */

/* [q counter VAL FROM TO STEP] -- count from FROM to TO by STEP, starting at
   initial value VAL. Defaults are: VAL = FROM = 0, TO = inf (infinity), STEP
   = 1. The counter wraps around at the end of the range. STEP can also be
   negative, and FROM>TO, to count backwards. The first outlet is the current
   counter value, the second outlet sends a bang when wrapping around. The
   inlets can be used to change the current value of the counter and the FROM,
   TO and STEP values at runtime. The first inlet also responds to the
   messages `bang' (next counter value) and `reset' (reset the counter to the
   FROM value). */

public counter;
private counterf VAL FROM TO STEP MSG;

counter			= obj (counterf,4,2;ref 0,ref 0,ref inf,ref 1);

counterf VAL FROM TO STEP (0,bang)
			= if sgn XSTEP*(XTO-XVAL) >= 0 then
			    (put VAL (XVAL+XSTEP) || (0,XVAL))
			  else
			    (put VAL (XFROM+XSTEP) || (1,bang;0,XFROM))
			  where [XVAL,XFROM,XTO,XSTEP] = map get
			    [VAL,FROM,TO,STEP];

counterf VAL FROM TO STEP (0,reset)
			= put VAL (get FROM);

counterf VAL FROM TO STEP (0,float X:Num)
			= put VAL X;
counterf VAL FROM TO STEP (1,float X:Num)
			= put FROM X;
counterf VAL FROM TO STEP (2,float X:Num)
			= put TO X;
counterf VAL FROM TO STEP (3,float X:Num)
			= put STEP X;

/* Stateless counter. Same as above, but implemented in a more functional
   style, using stream processing instead of mutable state. */

/* [q scounter VAL FROM TO STEP] -- see counter object above for a description
   of the parameters */

public scounter;
private scounterf VAL FROM TO STEP MSGS;

scounter	= stream_obj (scounterf,4,2;0,0,inf,1);

scounterf VAL FROM TO STEP {(0,bang)|MSGS}
		= {(0,VAL)|scounterf (VAL+STEP) FROM TO STEP MSGS}
		    if sgn STEP*(TO-VAL) >= 0;
		= {(1,bang),(0,FROM)|scounterf (FROM+STEP) FROM TO STEP MSGS}
		    otherwise;
scounterf VAL FROM TO STEP {(0,reset)|MSGS}
		= scounterf FROM FROM TO STEP MSGS;
scounterf _ FROM TO STEP {(0,float VAL)|MSGS}
		= scounterf VAL FROM TO STEP MSGS;
scounterf VAL _ TO STEP {(1,float FROM)|MSGS}
		= scounterf VAL FROM TO STEP MSGS;
scounterf VAL FROM _ STEP {(2,float TO)|MSGS}
		= scounterf VAL FROM TO STEP MSGS;
scounterf VAL FROM TO _ {(3,float STEP)|MSGS}
		= scounterf VAL FROM TO STEP MSGS;
