
/* q.c: a Q external for Pd. Requires an installation of the Q interpreter,
   available at http://q-lang.sf.net. Copyright (c) 2006 by Albert Graef,
   Johannes Gutenberg University Mainz, Germany. Distributed under the GNU
   General Public License (GPL) Version 2 or later. */

/* $Id: q.c,v 1.8 2007/10/05 23:53:48 agraef Exp $ */

/* This file is part of the Q programming system.

   The Q programming system is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   The Q programming system is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef _MSC_VER
#include <unistd.h>
#endif
#include <pthread.h>
#include <qint.h>
#include "m_pd.h"

#ifdef _MSC_VER
#define inline
#endif

/* extended error checking */
/* #define CHECK_SYNTAX 1 */
/* #define CHECK_OUTLET 1 */

/* The default QPATH, adapt this as needed. */

#ifndef QPATH
/* #define QPATH ".:/usr/share/q/lib:/usr/lib/q" */
#define QPATH ".:/usr/local/share/q/lib:/usr/local/lib/q"
#endif

/* Timer interval for polling asynchronous messages, in milliseconds. */
#define TIMER_INTERVAL 1

/* Ticks per millisecond of the internal clock. FIXME: Currently this is a
   fixed value (32.*441000./1000., cf. m_sched.c), will have to be rewritten
   if this changes. */
#define TICKS 14112.0

static char *q_getpath(void)
{
  char *s = getenv("QPATH");
  if (s)
    return s;
  else
    return QPATH;
}

static void q_setpath(char *path)
{
  char *s = malloc(strlen("QPATH=")+strlen(path)+1);
  if (s) {
    sprintf(s, "QPATH=%s", path);
    putenv(s);
  }
}

/* The Q class data structure. */

static t_class *q_class;

typedef struct _q {
  t_object x_obj;
  pthread_mutex_t mut_init;
  pthread_cond_t cond_init;
  int init;               /* set when initialized */
  int n_in, n_out;        /* number of extra inlets and outlets */
  struct _px **in;	  /* extra inlet proxies, see t_px below */
  t_outlet **out;	  /* outlets */
  qexpr foo;		  /* the object function */
  char *args;		  /* creation arguments */
  char *dir, *script;	  /* patch dir and script name ("q" receiver only) */
  struct _q *next, *prev; /* double-linked list of all q objects */
  qexpr *xv;              /* expression buffer for asynchronous messaging */
  int a_xv, n_xv;	  /* allocated and current buffer size */
  t_clock *x_clock;       /* wakeup for asynchronous processing */
  pthread_mutex_t mut_x;  /* mutex for protecting access to internal state */
  pthread_mutex_t mut_time;
  pthread_mutex_t mut_wait;
  pthread_cond_t cond_wait;
  double current_time;	  /* current (logical) Pd time */
  double wait_time;	  /* wait for specified time (pd_wait) */
  int wait_set;		  /* indicates that a new pd_wait is pending */
  t_clock *wait_clock;    /* wakeup for pending pd_wait */
} t_q;

/* Proxy objects for extra inlets (pilfered from flext code, thanks to Thomas
   Grill). */

static t_class *px_class;

typedef struct _px {
  t_object obj;
  t_q *x;		  /* parent */
  int ix;		  /* inlet index */
} t_px;

/* Keep track of a special "script" object and the list of all instantiated Q
   objects. The former is created by the setup routine and bound to the "q"
   symbol which accepts a filename symbol as a message to reload the
   script. The latter is provided so that we can traverse all Q objects and
   reinitialize them when reloading the script (see the q_init/q_fini routines
   below). */

static t_q *xscript = NULL, *xhead = NULL;
static int retval = 0;

/* Start the Q interpreter and load a script. The script is named "qext.q" by
   default and is searched for in the given directory dir before the normal Q
   path. */

static void q_init(t_q *x);
static void q_fini(t_q *x);

static void q_startup(char *dir, char *name)
{
  /* augment the Q path so that the directory of the patch is searched
     first */
  static const char *qpath0 = NULL;
  char *qpath;
  t_q *x;
  if (!qpath0) qpath0 = q_getpath();
  if (!dir || !*dir) dir = ".";
  if (!name || !*name) name = "pdext.q";
  qpath = malloc(strlen(dir)+strlen(qpath0)+2);
  if (qpath) {
    sprintf(qpath, "%s:%s", dir, qpath0);
    q_setpath(qpath);
    free(qpath);
  }
  /* get rid of the current object functions */
  for (x = xhead; x; x = x->next) q_fini(x);
  /* load the new script; if this fails we keep the previous script or use an
     empty script by default, so that the interpreter is properly initialized
     in any case */
  if ((retval = qexecl(name, 0))) {
    error("[q] error loading script '%s'", name);
    if ((retval = qexecl(NULL, 0))) {
      qrelease();
      return;
    }
  } else
    post("[q] loaded script '%s'", name);
  /* recompute the object functions */
  for (x = xhead; x; x = x->next) q_init(x);
  qrelease();
}

/* Helper functions to convert between Pd atoms and Q expressions. */

static char *get_expr(int argc, t_atom *argv)
{
  t_binbuf *b;
  char *exp_string, *s, *t;
  int exp_strlen, i;

  b = binbuf_new();
  binbuf_add(b, argc, argv);
  binbuf_gettext(b, &exp_string, &exp_strlen);
  exp_string = (char *)t_resizebytes(exp_string, exp_strlen, exp_strlen+1);
  exp_string[exp_strlen] = 0;
  if (!(s = malloc(exp_strlen+1)))
    return NULL;
  for (t = s, i = 0; i < exp_strlen; i++)
    if (exp_string[i] != '\\')
      *(t++) = exp_string[i];
  *t = 0;
  binbuf_free(b);
  freebytes(exp_string, exp_strlen+1);
  if ((t = realloc(s, strlen(s)+1)))
    s = t;
  return s;
}

static inline qexpr parse_expr(t_q *x, const char *s)
{
  int status;
  qexpr y = qparse(s, &status);
  if (y == NULL)
#if CHECK_SYNTAX
    /* complain about bad Q syntax */
    pd_error(x, "[q] invalid argument '%s'", s);
#else
    /* cast to a Q string */
    y = qmkstr(strdup(s));
#endif
  return y;
}

static inline void create_atom(t_atom *a, qexpr x)
{
  char *sval;
  double fval;
  if (qisstr(x, &sval)) {
    t_symbol *s = gensym(sval);
    SETSYMBOL(a, s);
  } else if (qisfloat(x, &fval) || qismpz_float(x, &fval))
    SETFLOAT(a, fval);
  else {
    int status;
    char *p = qprint(x, &status);
    t_symbol *s = gensym(p?p:"");
    SETSYMBOL(a, s);
    if (p) free(p);
  }
}

/* Initialize the object function (foo element) of a Q object, which is the
   function to be applied when a Q object is activated. This needs to be
   redone for all instantiated Q objects each time the script is reloaded. */

static char *argv_sym = NULL;

static void q_init(t_q *x)
{
  qexpr f, g, h, y, z, *xv;
  int n, status;
  long n_in = 1, n_out = 1;
  /* Set up a timer to poll for asynchronous messages. This should really be
     handled in worker threads, but there doesn't seem to be a safe way to
     schedule callbacks from secondary threads in the current stable Pd
     version. */
  clock_delay(x->x_clock, TIMER_INTERVAL);
  if (retval || x->args == NULL) return;
  f = qmkapp(qmksym(qsym(pd::varobj)), parse_expr(x, x->args));
  if (!f) return;
  f = qnewref(f);
  /* temporarily set the global PDOBJ variable to the Pd object being
     created */
  qdef("pd::PDOBJ", qmkobj(qtype(pd::PdObj), x));
  /* the current logical Pd time */
  x->current_time = clock_getlogicaltime()/TICKS;
  x->foo = qnewref(qevalx(f, &status));
  qdef("pd::PDOBJ", NULL);
  qfreeref(f); f = x->foo;
  if (!f || status != 0) {
    if (f) {
      int stat;
      char *p = qprint(f, &stat);
      pd_error(x, "[q] %s: %s", qstrerror(status), p);
      if (p) free(p);
    } else
      pd_error(x, "[q] %s", qstrerror(status));
    qfreeref(f);
    x->foo = NULL;
    return;
  }
  /* handle custom inlet/outlet configurations */
  if (qistuple(f, &n, &xv) && n > 0 && n <= 3 &&
      (n <= 1 || qisint(xv[1], &n_in)) &&
      (n <= 2 || qisint(xv[2], &n_out))) {
    x->foo = qnewref(xv[0]); qfreeref(f);
    if (n_in < 1) {
      pd_error(x, "[q] bad number %d of inlets, must be >= 1", (int)n_in);
      n_in = 1;
    }
    if (n_out < 0) {
      pd_error(x, "[q] bad number %d of outlets, must be >= 0", (int)n_out);
      n_out = 0;
    }
    x->n_in = n_in-1;
    x->n_out = n_out;
  }
}

static void q_fini(t_q *x)
{
  int i;
  pthread_mutex_lock(&x->mut_x);
  qfreeref(x->foo);
  x->foo = NULL;
  for (i = 0; i < x->n_xv; i++)
    qfreeref(x->xv[i]);
  if (x->xv) {
    free(x->xv);
    x->xv = NULL;
  }
  x->a_xv = x->n_xv = 0;
  pthread_mutex_unlock(&x->mut_x);
}

/* Process an output message and route it through the given outlet. */

static inline int check_outlet(t_q *x, int k)
{
  /* check outlet index */
#if CHECK_OUTLET
  /* complain about bad outlets */
  if (k < 0) {
    pd_error(x, "[q] bad outlet index %d, must be >= 0", k);
    return 0;
  } else if (x->n_out <= 0) {
    pd_error(x, "[q] bad outlet index %d, object has no outlets", k);
    return 0;
  } else if (k >= x->n_out) {
    pd_error(x, "[q] bad outlet index %d, must be < %d", k, x->n_out);
    return 0;
  } else
    return 1;
#else
  return (k >= 0) && (k < x->n_out);
#endif
}

static void send_message(t_q *x, int k, qexpr y)
{
  qexpr z = y, f, g = y;
  char *sval;
  double fval;
  int argc = 0, status, i;
  qexpr *args = NULL;
  t_atom *argv = NULL;
  /* get arguments */
  while (qisapp(g, &f, &y)) g = f, ++argc;
  if (argc > 0 &&
      !(args = malloc(argc*sizeof(qexpr))))
    return;
  g = z; i = argc;
  while (qisapp(g, &f, &y)) args[--i] = y, g = f;
  /* pd message generation */
  if (qisstr(g, &sval)) {
    if (!check_outlet(x, k)) goto errexit;
    if (argc == 0)
      /* single symbol value */
      outlet_anything(x->out[k], gensym(sval), 0, NULL);
    else {
      t_symbol *t = gensym(sval);
      if (argc > 0) {
	argv = malloc(argc*sizeof(t_atom));
	if (!argv) goto errexit;
      }
      for (i = 0; i < argc; i++)
	create_atom(&argv[i], args[i]);
      outlet_anything(x->out[k], t, argc, argv);
    }
  } else if (qisfloat(g, &fval) || qismpz_float(g, &fval)) {
    if (!check_outlet(x, k)) goto errexit;
    if (argc == 0)
      /* single float value */
      outlet_float(x->out[k], fval);
    else {
      /* create a list message with the float value in front */
      argv = malloc((argc+1)*sizeof(t_atom));
      if (!argv) goto errexit;
      create_atom(&argv[0], g);
      for (i = 0; i < argc; i++)
	create_atom(&argv[i+1], args[i]);
      outlet_list(x->out[k], &s_list, argc+1, argv);
    }
  } else {
    int sym = qexprsym(g);
    if (sym > 0 && sym != qvoidsym && sym != qnilsym) {
      /* manufacture a message with the given symbol as selector */
      char *pname;
      if (!check_outlet(x, k)) goto errexit;
      if ((pname = qprint(g, &status))) {
	t_symbol *t = gensym(pname);
	free(pname);
	if (argc > 0) {
	  argv = malloc(argc*sizeof(t_atom));
	  if (!argv) goto errexit;
	}
	for (i = 0; i < argc; i++)
	  create_atom(&argv[i], args[i]);
	outlet_anything(x->out[k], t, argc, argv);
      } else
	pd_error(x, "[q] memory allocation failed");
    }
  }
 errexit:
  if (args) free(args);
  if (argv) free(argv);
}

/* Handle a message to the given inlet. The message is first converted to a
   corresponding Q expression to which the object function is applied. We then
   evaluate the resulting expression in the Q interpreter and process the
   returned value, converting results back to corresponding Pd messages and
   sending these messages through the appropriate outlets. */

static void q_method(t_q *x, t_symbol *s, int k, int argc, t_atom *argv)
{
  int i;
  qexpr f = x->foo, y;

  /* check whether we have something to evaluate */
  if (!f) return;

  /* build the parameter expression from the message */
  qacquire();
  y = parse_expr(x, s->s_name);
  for (i = 0; i < argc && y != NULL; i++) {
    char buf[MAXPDSTRING];
    atom_string(argv+i, buf, MAXPDSTRING);
    y = qmkapp(y, parse_expr(x, buf));
  }
  /* add the inlet index */
  if (x->n_in > 0)
    y = qmktuplel(2, qmkint(k), y);
  if (y) {
    int status;
    /* evaluate */
    x->current_time = clock_getlogicaltime()/TICKS;
    y = qnewref(qevalx(qmkapp(f, y), &status));
    /* process the results and route them through the appropriate outlets */
    if (y && status == 0) {
      int n, m;
      qexpr *xv, *yv, hd, tl;
      long ix;
      qrelease();
      if (qiscons(y, &hd, &tl)) {
	qexpr z;
	for (z = y; qiscons(z, &hd, &tl); z = tl)
	  if (qistuple(hd, &m, &yv) &&
	      m == 2 && qisint(yv[0], &ix))
	    send_message(x, ix, yv[1]);
	  else
	    send_message(x, 0, hd);
      } else if (!qistuple(y, &n, &xv))
	send_message(x, 0, y);
      else if (n == 2 && qisint(xv[0], &ix))
	send_message(x, ix, xv[1]);
      else {
	for (i = 0; i < n; i++)
	  if (qistuple(xv[i], &m, &yv) &&
	      m == 2 && qisint(yv[0], &ix))
	    send_message(x, ix, yv[1]);
	  else
	    send_message(x, 0, xv[i]);
      }
      qacquire();
      qfreeref(y);
      qrelease();
    } else {
      if (y) {
	int stat;
	char *p = qprint(y, &stat);
	pd_error(x, "[q] %s: %s", qstrerror(status), p);
	if (p) free(p);
      } else
	pd_error(x, "[q] %s", qstrerror(status));
      qfreeref(y);
      qrelease();
    }
  } else
    qrelease();
}

/* Handle a message to the leftmost inlet. */

static void q_any(t_q *x, t_symbol *s, int argc, t_atom *argv)
{
  if (x->script) {
    /* special case of the "script" receiver object: load the given script
       (extra arguments are ignored right now) */
    free(x->script);
    x->script = strdup(s->s_name);
    qacquire();
    q_startup(x->dir, x->script);
  } else
    q_method(x, s, 0, argc, argv);
}

/* Handle messages to secondary inlets (these are routed through proxies,
   since only the first inlet of a Pd object can actually process arbitrary
   input messages). */

static void px_any(t_px *px, t_symbol *s, int argc, t_atom *argv)
{
  q_method(px->x, s, px->ix, argc, argv);
}

/* Timer for asynchronous message passing. */

static void q_tick(t_q *x)  
{
  int i, k, n, m;
  qexpr *xv, *yv, hd, tl;
  long ix;
  /* update the current time */
  pthread_mutex_lock(&x->mut_time);
  x->current_time = clock_getlogicaltime()/TICKS;
  pthread_mutex_unlock(&x->mut_time);
  /* process pending pd_wait */
  pthread_mutex_lock(&x->mut_wait);
  if (x->wait_set) {
    double systime = clock_getlogicaltime(),
      waittime = x->wait_time*TICKS;
    x->wait_set = 0;
    if (waittime <= systime) {
      x->wait_time = 0.0;
      pthread_cond_signal(&x->cond_wait);
    } else
      clock_set(x->wait_clock, waittime);
  }
  pthread_mutex_unlock(&x->mut_wait);
  /* process pending messages */
  pthread_mutex_lock(&x->mut_x);
  if (x->n_xv > 0) {
    for (k = 0; k < x->n_xv; k++) {
      qexpr y = x->xv[k];
      if (qiscons(y, &hd, &tl)) {
	qexpr z;
	for (z = y; qiscons(z, &hd, &tl); z = tl)
	  if (qistuple(hd, &m, &yv) &&
	      m == 2 && qisint(yv[0], &ix))
	    send_message(x, ix, yv[1]);
	  else
	    send_message(x, 0, hd);
      } else if (!qistuple(y, &n, &xv))
	send_message(x, 0, y);
      else if (n == 2 && qisint(xv[0], &ix))
	send_message(x, ix, xv[1]);
      else {
	for (i = 0; i < n; i++)
	  if (qistuple(xv[i], &m, &yv) &&
	      m == 2 && qisint(yv[0], &ix))
	    send_message(x, ix, yv[1]);
	  else
	    send_message(x, 0, xv[i]);
      }
    }
    qacquire();
    for (k = 0; k < x->n_xv; k++)
      qfreeref(x->xv[k]);
    qrelease();
    x->n_xv = 0;
  }
  pthread_mutex_unlock(&x->mut_x);
  clock_delay(x->x_clock, TIMER_INTERVAL);
}

/* Timer for pd_wait. */

static void q_wait(t_q *x)  
{
  pthread_mutex_lock(&x->mut_wait);
  x->wait_time = 0.0;
  pthread_cond_signal(&x->cond_wait);
  pthread_mutex_unlock(&x->mut_wait);
}

/* Create a new Q object. */

static void *q_new(t_symbol *s, int argc, t_atom *argv)
{
  int i;
  t_q *x = (t_q*)pd_new(q_class);

  x->next = xhead; x->prev = NULL;
  if (xhead) xhead->prev = x; xhead = x;
  x->init = 0;
  x->foo = NULL;
  x->dir = x->script = NULL;
  x->n_in = 0; x->n_out = 1;
  x->in = NULL;
  x->out = NULL;
  x->xv = NULL;
  x->a_xv = x->n_xv = 0;
  x->wait_time = 0.0;
  pthread_mutex_init(&x->mut_init, NULL);
  pthread_cond_init(&x->cond_init, NULL);
  pthread_mutex_init(&x->mut_time, NULL);
  pthread_mutex_init(&x->mut_wait, NULL);
  pthread_cond_init(&x->cond_wait, NULL);
  pthread_mutex_init(&x->mut_x, NULL);
  x->x_clock = clock_new(x, (t_method)q_tick);
  x->wait_clock = clock_new(x, (t_method)q_wait);
  if (argc <= 0) {
    x->args = NULL;
    return (void *)x;
  } else {
    x->args = get_expr(argc, argv);
    if (!x->args) {
      pd_error(x, "[q] memory allocation failed");
      return (void *)x;
    }
  }
  pthread_mutex_lock(&x->mut_x);
  /* initialize the object function and determine the number of inlets and
     outlets (these cannot be changed later) */
  if (x->args != NULL) {
    qacquire();
    q_init(x);
    qrelease();
  }
  if (x->foo != NULL) {
    /* allocate memory for inlets and outlets */
    if (x->n_in > 0)
      x->in = malloc(x->n_in*sizeof(t_px*));
    if (x->n_out > 0)
      x->out = malloc(x->n_out*sizeof(t_outlet*));
    if (x->n_in > 0 && x->in == NULL ||
	x->n_out > 0 && x->out == NULL)
      pd_error(x, "[q] memory allocation failed");
    if (!x->in) x->n_in = 0;
    if (!x->out) x->n_out = 0;
    /* initialize the proxies for the extra inlets */
    for (i = 0; i < x->n_in; i++) {
      x->in[i] = (t_px*)pd_new(px_class);
      x->in[i]->x = x;
      x->in[i]->ix = i+1;
      inlet_new(&x->x_obj, &x->in[i]->obj.ob_pd, 0, 0);
    }
    /* initialize the outlets */
    for (i = 0; i < x->n_out; i++)
      x->out[i] = outlet_new(&x->x_obj, 0);
  } else {
    x->n_in = x->n_out = 0;
  }
  pthread_mutex_unlock(&x->mut_x);
  pthread_mutex_lock(&x->mut_init);
  x->init = 1;
  pthread_cond_broadcast(&x->cond_init);
  pthread_mutex_unlock(&x->mut_init);
  return (void *)x;
}

/* Finalize a Q object. */

static void q_free(t_q *x)
{
  int i;
  clock_free(x->x_clock);
  free(x->args);
  qacquire();
  q_fini(x);
  qrelease();
  pthread_cond_destroy(&x->cond_init);
  pthread_mutex_destroy(&x->mut_init);
  pthread_mutex_destroy(&x->mut_time);
  pthread_cond_destroy(&x->cond_wait);
  pthread_mutex_destroy(&x->mut_wait);
  pthread_mutex_destroy(&x->mut_x);
  if (x->dir) free(x->dir);
  if (x->script) free(x->script);
  for (i = 0; i < x->n_in; i++)
    pd_free((t_pd*)x->in[i]);
  if (x->in) free(x->in);
  if (x->out) free(x->out);
  if (x->prev)
    x->prev->next = x->next;
  else
    xhead = x->next;
  if (x->next)
    x->next->prev = x->prev;
}

/* Setup routine for this plugin. */

void q_setup(void) {
  t_symbol *q_s = gensym("q");
  char *dir = canvas_getcurrentdir()->s_name;

  /* proxies for extra inlets */
  px_class = class_new(gensym("q proxy"),
		       NULL, NULL,
		       sizeof(t_px), CLASS_PD|CLASS_NOINLET,
		       A_NULL);
  class_addanything(px_class, px_any);

  /* q objects */
  q_class = class_new(q_s,
		      (t_newmethod)q_new, (t_method)q_free,
		      sizeof(t_q), CLASS_DEFAULT,
		      A_GIMME, A_NULL);
  class_addanything(q_class, q_any);
  class_sethelpsymbol(q_class, gensym("q-help"));

  /* create a receiver for the q script message */
  xscript = (t_q*)q_new(q_s, 0, NULL);
  if (dir)
    xscript->dir = strdup(dir);
  xscript->script = strdup("pdext.q");
  pd_bind(&xscript->x_obj.ob_pd, q_s);

  /* start up the Q interpreter */
  q_startup(xscript->dir, xscript->script);
  qdef("pd::PDGLOB", qmkobj(qtype(pd::PdObj), xscript));
}
